;;; $Id: contlib.scm,v 1.11 1993/12/08 18:01:14 queinnec Exp $

;;;                 The code of the associated paper
;;;         presenting most of the current control operators.
;;;                        Christian Queinnec
;;;           \'Ecole Polytechnique & INRIA-Rocquencourt

;;; This code might seem to have a strange architecture. The reason is
;;; that it is made from several different files from different
;;; systems or book or papers just concatenated here. The overall
;;; architecture is the following:
;;;        A naive interpreter in direct style
;;;        Its associated direct style library.
;;;        An ACPS transformer.
;;;        A macro expander
;;;        syntaxes associated to the control operators
;;;        An ACPS converted library
;;;        The library of control operators
;;;        Some testing functions.

;;;=========================================================================
;;;=========================================================================
;;; A simple interpreter: the reason not to use the underlying eval is 
;;; to be able to easily catch errors.
;;; This part is snarfed from ~/DEA/natScheme.scm
;;; extend has been changed to r.extend
;;; *left-to-right* has been added. In fact it has no interest here
;;; since the order is imposed by ACPS (see below).

;;; A native Scheme in Scheme
;;; Scheme written in native (and unrestricted) Scheme.
;;; progn is implicit in LAMBDA.

(define *left-to-right* #t)

(define (evaluate e env)
  (if (atom? e) 
      (cond ((symbol? e) (lookup e env))
            ((or (number? e) (string? e) (char? e) (boolean? e) (vector? e)) 
             e )
            (else (wrong "Cannot evaluate" e)) )
      (case (car e)
        ;; No error checking on special forms
        ((quote)  (cadr e))
        ((if)     (if (evaluate (cadr e) env)
                      (evaluate (caddr e) env)
                      (evaluate (cadddr e) env) ))
        ;; begin is not really a special form
        ((begin)  (eprogn (cdr e) env))
        ((set!)   (update! (cadr e) env (evaluate (caddr e) env)))
        ((lambda) (make-function (cadr e) (cddr e) env))
        (else     (if *left-to-right*
                      (let ((fn (evaluate (car e) env)))
                        (invoke fn (evlis (cdr e) env)) )
                      (let ((args (evlis (cdr e) env)))
                        (invoke (evaluate (car e) env) args) ))) ) ) )

;;; Two specialized evaluators: 

(define (evlis exps env)
  (if (pair? exps)
      (if *left-to-right*
          (let ((val (evaluate (car exps) env)))
            (cons val (evlis (cdr exps) env)) )
          (let ((vals (evlis (cdr exps) env)))
            (cons (evaluate (car exps) env) vals) ) )
      '() ) )

(define (eprogn exps env)
  (if (pair? exps)
      (if (pair? (cdr exps))
          (begin (evaluate (car exps) env)
                 (eprogn (cdr exps) env) )
          (evaluate (car exps) env) )
      '() ) )

;;; Functions are represented as functions

(define (make-function variables body env)
  (lambda (values)
     (eprogn body (r.extend env variables values)) ) )

(define (invoke fn args)
  (if (procedure? fn) 
      (fn args)
      (wrong "Not a function" fn) ) )

;;; Environments are Alists handled by lookup and update!.
;;; The global environment is not exensible (see definitial below).

(define (lookup id env)
  (if (pair? env)
      (if (eq? (caar env) id)
          (cdar env)
          (lookup id (cdr env)) )
      (wrong "No such binding" id) ) )

(define (update! id env value)
  (if (pair? env)
      (if (eq? (caar env) id)
          (begin (set-cdr! (car env) value)
                 ;; Returns value (one of the possible choice)
                 value )
          (update! id (cdr env) value) )
      (wrong "No such binding" id) ) )

(define (r.extend env names values)
  (cond ((pair? names)
         (if (pair? values)
             (cons (cons (car names) (car values))
                   (r.extend env (cdr names) (cdr values)) )
             (wrong "Too less values") ) )
        ((null? names)
             (if (null? values)
                 env 
                 (wrong "Too much values") ) )
        ((symbol? names) (cons (cons names values) env)) ) )

(define r.init '())

(define r.global r.init)

;;;                    [[ SLIB macro ]]
;;; If you run this file with SLIB, then comment out the following:

;(defmacro define-pervasive-macro (call . body)
;  `(defmacro ,(car call) ,(cdr call) . ,body) )

;;; Three macros to define the initial global environment

(define-pervasive-macro (definitial name value)
  `(begin (set! r.global (cons (cons ',name ,value) r.global))
          ',name ) )

(define-pervasive-macro (defprimitive name value arity)
  `(definitial ,name 
     (lambda (values) 
       (if (= ,arity (length values))
           (apply ,value values)
           (wrong "Incorrect arity"
                  (list ',name values) ) ) ) ) )

(define the-false-value #f)

(define-pervasive-macro (defpredicate name value arity)
  `(definitial ,name
     (lambda (values) 
       (if (= ,arity (length values))
           (or (apply ,value values) the-false-value)
           (wrong "Incorrect arity"
                  (list ',name values) ) ) ) ) )

(defprimitive cons cons 2)
(defprimitive car car 1)
(defprimitive cdr cdr 1)
(defpredicate pair? pair? 1)
(defpredicate eq? eq? 2)
(defprimitive set-car! set-car! 2)
(defprimitive set-cdr! set-cdr! 2)
(defprimitive + + 2)
(defprimitive - - 2 )
(defprimitive = = 2)
(defpredicate < < 2)
(defpredicate > > 2)
(defprimitive * * 2)
(defpredicate <= <= 2)
(defpredicate >= >= 2)
(defprimitive remainder remainder 2)
(defprimitive reverse reverse 1)
(defprimitive append append 2)
(defprimitive assq assq 2)
(defprimitive not not 1)
(defprimitive caar caar 1)
(defprimitive cdar cdar 1)
(defprimitive equal? equal? 2)

(define (make-list n item)
  (if (> n 0) (cons item (make-list (- n 1) item)) '()) )
(defprimitive make-list make-list 2)

(defprimitive apply invoke 2)

;;; Some fundamental constants:
(definitial t #t)
(definitial f #f)
(definitial nil '())

(definitial list
   (lambda (values) values) )

;;; necessary if they occur in interp-define.
(defprimitive show show 1)       ;; DEBUG
(defprimitive newline newline 0) ;; DEBUG
(defprimitive display display 1) ;; DEBUG

;;;=========================================================================
;;;=========================================================================
;;; ACPS: hygien is hand-made with gensym and symbol-append. 
;;; This part is snarfed and adapted from ~/icslas/pwb-v.scm

;;; Note that global variables are handled specially not to confuse 
;;; libraries with or without ACPS. 
;;; E is the expression to convert, R the environment (the list of local
;;; variables) and Q the representation of the continuation.

(define (acps e r q)
  (if (atom? e) 
      (if (symbol? e)
          (acps-reference e r q)
          `(resume ,q ',e) )
      (case (car e)
        ((quote)  `(resume ,q ,e))
        ((if)     (acps-if (cadr e) (caddr e) (cadddr e) r q))
        ((begin)  (acps-begin (cdr e) r q))
        ((set!)   (acps-set! (cadr e) (caddr e) r q))
        ((lambda) (acps-lambda (cadr e) (cddr e) r q))
        (else     (acps-application e r q)) ) ) )

(define (acps-reference n r q)
  (let ((n (if (memq n r) n (symbol-append 'acps- n))) )
    `(resume ,q ,n) ) )

(define (acps-set! n e r q)
  (acps e 
        r
        `(extend
          ,q (make-frame 
              ,(let ((v (gensym "Value"))
                     (q (gensym "Q"))
                     (n (if (memq n r) n (symbol-append 'acps- n))) )
                 `(lambda (,q ,v)
                    (resume ,q (set! ,n ,v)) ) ) ) ) ) )

(define (acps-if p c a r q)
  (acps p 
        r
        `(extend
          ,q (make-frame
              ,(let ((b (gensym "Bool"))
                     (q (gensym "Q")) )
                 `(lambda (,q ,b) 
                    (if ,b ,(acps c r q) 
                        ,(acps a r q) ) ) ) ) ) ) )

(define (acps-begin e* r q)
  (if (pair? (cdr e*))
      (acps (car e*)
            r
            `(extend
              ,q (make-frame
                  ,(let ((void (gensym "void"))
                         (q (gensym "Q")) )
                     `(lambda (,q ,void)
                        ,(acps-begin (cdr e*) r q) ) ) ) ) )
      (acps (car e*) r q) ) )

(define (acps-lambda n* e* r q)
  `(resume ,q
           ,(let ((q (gensym "Q")))
              `(lambda (,q . ,n*)
                 ,(acps-begin e* (append (flat n*) r) q) ) ) ) )

(define (acps-application ee* r q)
  (let ((vv* (map (lambda (e) (gensym "V")) ee*))
        (q* (map (lambda (e) (gensym "Q")) ee*)) )
    (define (evlis e* v* q q*)
      (acps (car e*) 
            r
            `(extend
              ,q (make-frame
                  (lambda (,(car q*) ,(car v*)) 
                    ,(if (pair? (cdr e*))
                         (evlis (cdr e*) (cdr v*) 
                                (car q*) (cdr q*) )
                         `(,(car vv*) ,(car q*) . ,(cdr vv*) ) ) ) ) ) ) )
    (if *left-to-right* 
        (evlis ee* vv* q q*)
        (evlis (reverse ee*) (reverse vv*) q q*) ) ) )

;;;=========================================================================
;;;=========================================================================
;;; Macro expansion
;;; snarfed from ~/approf/expand.scm

;;; copied from XLA/expand.scm
;;; Lacks some features on cond and some other syntaxes (delay)
;;; Syntax errors are not harmonious (use match-lambda).

;;;===================================================================
;;;                   Expansion Passing Style
;;; After R. Kent Dybvig, Daniel P. Friedman and Christopher T. Haynes
;;; Lisp and Symbolic Computation, 1, 53-75 (1988),
;;; Kluwer Academic Publishers.

;;; Christian Queinnec  queinnec@poly.polytechnique.fr
;;; \'Ecole Polytechnique \& INRIA--Rocquencourt
;;; Ecole Polytechnique, 91128 Palaiseau Cedex, France
;;;===================================================================

;;; The entry point used by toplevel4
(define (%expand form current-macroexpand)
  (macro-walk form current-macroexpand) ) 

;;; The entry point is (expand-syntax form),
;;; New expanders may be defined by (define-expander ...), 
;;; they are added to current-macroexpand.
;;; initial-macroexpand corresponds to a syntax without macros.
;;; standard-macroexpand to the usual syntax of Scheme

(define (%expand-syntax form)
  (macro-walk form current-macroexpand) )

;;;===========================================================
;;; Internal guts

(define (macro-walk exp macroexpand)
  (macroexpand exp macroexpand) )

;;; Keep the last cdr of expressions to be macroexpanded.
(define (macro-walk* exps macroexpand)
  (if (pair? exps)
    (cons (macro-walk (car exps) macroexpand)
          (macro-walk* (cdr exps) macroexpand) )
    exps ) )

;;; The initial macroexpand. It just walks the expression and tries
;;; to macroexpand anything it recognizes. Note that no syntax is
;;; checked since it is possible to specify macros that overload
;;; the native syntax.
(define (initial-macroexpand exp macroexpand)
  (if (atom? exp) exp
      (case (car exp)
        ((quote)  exp)
        ((if)     `(if . ,(macro-walk* (cdr exp) macroexpand)))
        ((begin)  `(begin . ,(macro-walk* (cdr exp) macroexpand)))
        ((set!)   `(set! . ,(macro-walk* (cdr exp) macroexpand)))
        ((lambda) (expand-lambda exp macroexpand))
        ((with)   (expand-lambda exp macroexpand))
        (else     (macro-walk* exp macroexpand)) ) ) )

;;; Lambda (and with) are specific cases since local variables shadow 
;;; macros. But then it is necessary to recognize the variables !
(define (expand-lambda exp macroexpand)
  (if (pair? (cdr exp))
      `(,(car exp) ,(cadr exp)
                   . ,(macro-walk* 
                       (cddr exp)
                       (macro-variables-extend (cadr exp) macroexpand) ) )
      exp ) )

;;; Add to macroexpand that variables shadow macros whatever their
;;; position are (function or parameter). Check however if the list
;;; of variables is syntactically correct otherwise don't shadow.
(define (macro-variables-extend variables macroexpand)
  (define (check-variable-list o)
    (or (null? o)
        (symbol? o)
        (and (pair? o)
             (symbol? (car o))
             (check-variable-list (cdr o)) ) ) )
  (if (check-variable-list variables)
      (let ((variables (flat variables)))
        (macro-atom-extend* 
         (macro-form-extend* 
          macroexpand
          variables
          (lambda (exp2 macroexpand2)
            `(,(car exp2) . ,(macro-walk* (cdr exp2) macroexpand2)) ) )
         variables
         (lambda (exp2 macroexpand2)
           exp2 ) ) )
      macroexpand ) )

;;;===========================================================
;;; Two different extenders of syntax are provided. The first one
;;; recognizes forms with a given keyword in functional position, the
;;; other is triggered on single atoms.

;;; Add a new association between a key and an expander. If the
;;; key appears in the car of a form, the expander will be invoked.
(define (macro-form-extend macroexpand key expander)
  (lambda (exp2 macroexpand2)
      (if (atom? exp2) (macroexpand exp2 macroexpand2)
          (if (eq? (car exp2) key)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) ) ) ) )

;;; Add a common expander to many keys.
(define (macro-form-extend* macroexpand keys expander)
  (lambda (exp2 macroexpand2)
      (if (atom? exp2) (macroexpand exp2 macroexpand2)
          (if (member (car exp2) keys)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) ) ) ) )

;;; It is also possible to associate an expander to a single atom,
;;; something reminiscent of symbol-macrolet.
(define (macro-atom-extend macroexpand key expander)
  (lambda (exp2 macroexpand2)
    (if (atom? exp2) 
        (if (eq? exp2 key)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) )
        (macroexpand exp2 macroexpand2) ) ) )

;;; Add a common expander to many atoms
(define (macro-atom-extend* macroexpand keys expander)
  (lambda (exp2 macroexpand2)
    (if (atom? exp2) 
        (if (member exp2 keys)
            (expander exp2 macroexpand2)
            (macroexpand exp2 macroexpand2) )
        (macroexpand exp2 macroexpand2) ) ) )

(set! current-macroexpand initial-macroexpand)

;;;===========================================================

;;; Defines an expander and associate it under a name.
;;; macroexpand may be used in the body of expanders.
;;; The expanded result is not automatically expnded again, so
;;; macros defined with define-expander have to invoke macroexpand.
(define-pervasive-macro (define-expander call . body)
  (let ((name (car call))
        (variables (cdr call)) )
    (define (map-arguments variables parameters)
      (cond ((pair? variables)
             (cons `( ,(car variables) (car ,parameters) )
                   (map-arguments (cdr variables) `(cdr ,parameters)) ) )
            ((symbol? variables)
             (list `( ,variables ,parameters )) )
            ((null? variables) '()) ) )
    `(set! current-macroexpand
           (macro-form-extend 
            current-macroexpand
            ',name
            (lambda (c\@ll macroexpand)
              (let ,(map-arguments (cdr call) '(cdr c\@ll))
                . ,body ) ) ) ) ) )

;;;===========================================================
;;; Other expanders for derived expression types (DP1178 on Scheme)

(define-expander (let* binds . body)
  (macro-walk (if (pair? binds)
                  `(let (,(car binds))
                     (let* ,(cdr binds) . ,body) )
                  (handle-sequence body) )
              macroexpand ) )

(define (handle-sequence body)
  (if (pair? body)
      (if (null? (cdr body))
          (car body)
          `(begin . ,body) )
      (expand-error 'empty-begin body) ) )

(define-expander (let binds . body)
  (if (symbol? binds)
      ;; Named let
      (macro-walk 
       (let ((name binds)
             (binds (car body))
             (body (cdr body)) )
         (sort-bindings
          binds
          (lambda (inited uninited)
            (when (pair? uninited)
                  (expand-error 
                   "Named let is incompatible of uninitialized bindings" ) )
            `((letrec ((,name (lambda ,(map car binds)
                                . ,body )))
                ,name )
              . ,(map cadr binds) ) ) ) )
       macroexpand )
      ;; Anonymous let
      (sort-bindings 
       binds
       (lambda (inited uninited)
         (let* ((body (analyze-body body macroexpand))
                (body (if (pair? uninited)
                          `(with ,uninited ,body)
                          body )) )
           (if (pair? inited)
               `((lambda ,(map car inited) ,body)
                 . ,(macro-walk* (map cadr inited) macroexpand) )
               body ) ) ) ) ) )
             
;;; Sort the bindings of a let into two classes: these that are initialized
;;; and the others.
(define (sort-bindings bindings k)
  (let scan ((bindings bindings)
             (inited '())
             (uninited '()) )
    (if (pair? bindings)
        (if (pair? (car bindings))
            (scan (cdr bindings) (cons (car bindings) inited) uninited)
            (scan (cdr bindings) inited (cons (car bindings) uninited)) )
        (k (reverse! inited)
           (reverse! uninited) ) ) ) )

(define (analyze-body body macroexpand)
  (collect-definitions 
   body
   (lambda (definitions body)
     (macro-walk 
      (if (null? definitions)
          (handle-sequence body)
          (let* ((assignments 
                  (map (lambda (definition)
                         (analyze-definition (cadr definition)
                                             (cddr definition)
                                             macroexpand ) )
                       definitions ) )
                 (names (map cadr assignments)) )
            `(letrec ,(map cdr assignments)
               ,(handle-sequence body) ) ) )
      macroexpand ) ) ) )

;(define-expander (letrec binds . body)
;  (if (pair? binds)
;      `(with ,(map (lambda (bind) 
;                     (if (pair? bind) (car bind) bind) )
;                   binds )
;             ,@(map (lambda (bind) 
;                      (if (pair? bind) 
;                          (macro-walk `(set! . ,bind) macroexpand)
;                          `#t ) )
;                    binds )
;             ,(analyze-body body macroexpand) )
;      (analyze-body body macroexpand) ) )

(define-expander (letrec binds . body)
  (if (pair? binds)
      `((lambda ,(map (lambda (bind) 
                        (if (pair? bind) (car bind) bind) )
                      binds )
          ,@(map (lambda (bind) 
                   (if (pair? bind) 
                       (macro-walk `(set! . ,bind) macroexpand)
                       `#t ) )
                 binds )
          ,(analyze-body body macroexpand) )
        . ,(map (lambda (bind) 0) binds) )
      (analyze-body body macroexpand) ) )

(define-expander (define call . body)
  (analyze-definition call body macroexpand) )

(define (analyze-definition call body macroexpand)
  (if (pair? call)
      `(set! ,(car call)
             (lambda ,(cdr call)
               ,(analyze-body body macroexpand) ) )
      (if (and (pair? body)
               (null? (cdr body)) )
          `(set! ,call ,(macro-walk (car body) macroexpand))
          (expand-error 'incorrect-definition call) ) ) )

(define-expander (lambda variables . body)
  `(lambda ,variables 
     ,(analyze-body body (macro-variables-extend variables macroexpand)) ) )

(define (collect-definitions e* k)
  (if (and (pair? e*) (pair? (car e*)) (eq? (caar e*) 'define))
      (collect-definitions 
       (cdr e*)
       (lambda (bindings body)
         (k (cons (car e*) bindings) body) ) )
      (k '() e*) ) )

(define-expander (and . arguments)
  (if (pair? arguments)
      (macro-walk (if (pair? (cdr arguments))
                      `(if ,(car arguments)
                           (and . ,(cdr arguments))
                           #f )
                      (car arguments) )
                  macroexpand )
      '#t ) )

(define-expander (or . arguments)
  (if (pair? arguments)
      (macro-walk (if (pair? (cdr arguments))
                      (let ((g (gensym)))
                        `((lambda (,g)
                            (if ,g ,g (or . ,(cdr arguments))) )
                          ,(car arguments) ) )
                      (car arguments) )
                  macroexpand )
      '#f ) )

(define-expander (cond . arguments)
  (macro-walk
   (if (pair? arguments)
       (if (pair? (car arguments))
           (if (eq? (car (car arguments)) 'else)
               (if (null? (cdr arguments))
                   (handle-sequence (cdr (car arguments)))
                   (expand-error 'cond-else-not-last) )
               `(if ,(car (car arguments))
                    ,(handle-sequence (cdr (car arguments)))
                    (cond . ,(cdr arguments)) ) )
           (expand-error 'cond-1st-clause) )
       (expand-error 'no-clause-at-all) )
   macroexpand ) )

(define-expander (begin0 first . others)
  (macro-walk (let ((g (gensym)))
                `((lambda (,g) 
                    ,(handle-sequence others)
                    ,g )
                  ,first ) )
              macroexpand ) )

(define-expander (case exp . clauses)
  (let ((g (gensym)))
    (define (generate clauses)
      (if (pair? clauses)
          (if (pair? (car clauses))
              (if (eq? (caar clauses) 'else)
                  (if (null? (cdr clauses))
                      (handle-sequence (cdar clauses))
                      (expand-error 'case-else-not-last clauses) )
                  `(if (memq ,g ',(caar clauses))
                       ,(handle-sequence (cdar clauses))
                       ,(generate (cdr clauses)) ) )
              (expand-error 'case-incorrect-clause) )
          `*UNSPECIFIED* ) )
    (macro-walk `((lambda (,g) ,(generate clauses)) ,exp)
                macroexpand ) ) )

(define-expander (do bindings test . body)
  (define (get list n default)
    (if (> n 0)
        (if (pair? list)
            (get (cdr list) (- n 1) default)
            (expand-error 'incorrect-do-form list) )
        (if (pair? list) (car list) (default)) ) )
  (let ((g (gensym)))
    (macro-walk
     `(letrec ((,g (lambda ,(map car bindings)
                     (if ,(car test)
                         ,(handle-sequence (cdr test))
                         (begin (begin . ,body)
                                (,g . ,(map (lambda (bind)
                                              (get bind 2 (lambda () 
                                                            (car bind) )) )
                                            bindings )) ) ) )))
                                                     
        (,g . ,(map (lambda (bind)
                      (get bind 1 (lambda () 
                                    (expand-error
                                     'missing-init-in-do bind ) )) )
                    bindings )) )
     macroexpand ) ) )

;;; This quasiquotation is restricted to a single level
(define-expander (quasiquote exp)
  (define (quasiquote-walk e)
    (if (pair? e)
        (if (eq? (car e) 'unquote)
            (if (pair? (cdr e))
                (if (null? (cddr e))
                    (cadr e)
                    (expand-error incorrect-unquote-form e) )
                (expand-error incorrect-unquote-form e) )
            (if (eq? (car e) 'quasiquote)
                (expand-error "No embedded quasiquotation" e);; TEMP
                (quasiquote-pair-walk e) ) )
        (list 'quote e) ) )
  (define (quasiquote-pair-walk e)
    (if (pair? (car e))
        (if (eq? (car (car e)) 'unquote-splicing)
            (if (pair? (cdr (car e)))
                (if (null? (cddr (car e)))
                    (list 'append
                          (cadr (car e))
                          (quasiquote-walk (cdr e)) )
                    (expand-error incorrect-unquote-splicing-form (car e)) )
                (expand-error incorrect-unquote-splicing-form (car e)) )
            (list 'cons
                  (quasiquote-walk (car e))
                  (quasiquote-walk (cdr e)) ) )
        (list 'cons
              (list 'quote (car e))
              (quasiquote-walk (cdr e)) ) ) )
  (macro-walk (quasiquote-walk exp) macroexpand) )

;;;===========================================================
;;; These are not defined in Scheme R4RS:
(define-expander (when condition . body)
  (macro-walk `(if ,condition ,(handle-sequence body) '*UNSPECIFIED*)
              macroexpand ) )

(define-expander (unless condition . body)
  (macro-walk `(if ,condition '*UNSPECIFIED* ,(handle-sequence body))
              macroexpand ) )

;;;===========================================================
;;; The standard syntax is now the current one.

;(define standard-macroexpand current-macroexpand)


;;;==========================================================================
;;;==========================================================================
;;; Library of primitives

;;; Define appropriate syntaxes. These are first defined with
;;; define-syntax to be included (automatically with LiSP2TeX) in the
;;; paper, then they are defined with the appropriate expander to be
;;; recognized by the above interpreter.

(define-syntax prompt
  (syntax-rules ()
    ((_ expression) (prompt-evaluate (lambda () expression))) ) )

(define-syntax reset
  (syntax-rules ()
    ((_ expression) (reset-evaluate (lambda () expression))) ) )

(define-syntax shift
  (syntax-rules ()
    ((_ variable expression) 
     (shift-evaluate (lambda (variable) expression)) ) ) )

(define-expander (prompt expression)
  (macro-walk `(prompt-evaluate (lambda () ,expression))
              macroexpand ) )

(define-expander (reset expression)
  (macro-walk `(reset-evaluate (lambda () ,expression))
              macroexpand ) )

(define-expander (shift variable expression)
  (macro-walk `(shift-evaluate (lambda (,variable) ,expression))
              macroexpand ) )

;;; and make these syntaxes available
(define standard-macroexpand current-macroexpand)

;;;==========================================================================
;;; Define now the primitives of the library of the interpreter with
;;; the appropriate acps interface.

(define-pervasive-macro (defACPSprimitive name value arity)
  `(defprimitive ,(symbol-append 'ACPS- name)
     (let ((f ,value))
       (lambda (q . args)
         (resume q (apply f args)) ) )
     ,(+ 1 arity) ) )

(define-pervasive-macro (defACPSpredicate name value arity)
  `(defprimitive ,(symbol-append 'ACPS- name)
     (let ((f ,value))
       (lambda (q . args)
         (resume q (or (apply f args) the-false-value)) ) )
     ,(+ 1 arity) ) )

(defACPSprimitive cons cons 2)
(defACPSprimitive car car 1)
(defACPSprimitive cdr cdr 1)
(defACPSpredicate pair? pair? 1)
(defACPSpredicate eq? eq? 2)
(defACPSprimitive set-car! set-car! 2)
(defACPSprimitive set-cdr! set-cdr! 2)
(defACPSprimitive + + 2)
(defACPSprimitive - - 2 )
(defACPSprimitive = = 2)
(defACPSpredicate < < 2)
(defACPSpredicate > > 2)
(defACPSprimitive * * 2)
(defACPSpredicate <= <= 2)
(defACPSpredicate >= >= 2)
(defACPSprimitive remainder remainder 2)
(defACPSprimitive symbol? symbol? 1)
(defACPSprimitive null? null? 1)
(defACPSpredicate equal? equal? 2)

(definitial ACPS-t #t)
(definitial ACPS-f #f)
(definitial ACPS-nil '())

(definitial ACPS-apply
  (lambda (values)
    (if (>= (length values) 3)
        (let ((q (car values))
              (f (cadr values))
              (args (cddr values)) )
          (define (flat args)
            (if (pair? args)
                (if (pair? (cdr args))
                    (cons (car args) (flat (cdr args)))
                    (car args) )
                '() ) )
          (invoke f (cons q (flat args))) )
        (wrong "Incorrect arity for apply" values) ) ) )

(definitial ACPS-list
  (lambda (values)
    (if (>= (length values) 1)
        (resume (car values) (cdr values))
        (wrong "Incorrect arity for list" values) ) ) )

(defACPSprimitive assq assq 2)
(defACPSprimitive display display 1) 
(defACPSprimitive newline newline 0)
(defACPSprimitive show show 1)

(defACPSprimitive error wrong 1)

;;; Some free locations since the global environment cannot be extended
;;; dynamically.

(definitial ACPS-a 'void)
(definitial ACPS-foo 'void)
(definitial ACPS-fact 'void)
(definitial ACPS-primes 'void)
(definitial ACPS-multiply-list 'void)
(definitial ACPS-find-all 'void)
(definitial ACPS-//search 'void)
(definitial ACPS-times2list 'void)
(definitial ACPS-plus2list 'void)
(definitial ACPS-spawn/exit 'void)
(definitial ACPS-product0 'void)
(definitial ACPS-mapuntil 'void)
(definitial ACPS-stimes 'void)
(definitial ACPS-reverse 'void)
(definitial ACPS-iterate 'void)
(definitial ACPS-xpl 'void)
(definitial ACPS-xpl3 'void)
(definitial ACPS-xpl4 'void)
(definitial ACPS-getslice 'void)
(definitial ACPS-visit 'void)
(definitial ACPS-every? 'void)
(definitial ACPS-some? 'void)
(definitial ACPS-same-fringe 'void)
;(definitial ACPS- 'void)

;;; define all classes needed by control operators. 

;;;                  [[ What if you do not have Meroon ]]
;;; Just comment out the following bunch of definitions.

;(define frame-behavior 
;  (lambda (value)
;    (vector-ref value 1) ) )
;(defprimitive frame-behavior frame-behavior 1)
;(define make-frame
;  (lambda (behavior)
;    (vector 'frame behavior) ) )
;(defprimitive make-frame make-frame 1)
;(define frame?
;  (lambda (value)
;    (and (vector? value)
;         (eq? (vector-ref value 0) 'frame) ) ) )
;(defpredicate frame? frame? 1)
;(define make-prompt-frame
;  (lambda (behavior)
;    (vector 'prompt-frame behavior) ) )
;(defprimitive make-prompt-frame make-prompt-frame 1)
;(define prompt-frame?
;  (lambda (value)
;    (and (vector? value)
;         (eq? (vector-ref value 0) 'prompt-frame) ) ) )
;(defpredicate prompt-frame? prompt-frame? 1)
;(define make-reset-frame
;  (lambda (behavior)
;    (vector 'reset-frame behavior) ) )
;(defprimitive make-reset-frame make-reset-frame 1)
;(define reset-frame?
;  (lambda (value)
;    (and (vector? value)
;         (eq? (vector-ref value 0) 'reset-frame) ) ) )
;(defpredicate reset-frame? reset-frame? 1)
;(define make-spawn-frame
;  (lambda (behavior)
;    (vector 'spawn-frame behavior) ) )
;(defprimitive make-spawn-frame make-spawn-frame 1)
;(define spawn-frame?
;  (lambda (value)
;    (and (vector? value)
;         (eq? (vector-ref value 0) 'spawn-frame) ) ) )
;(defpredicate spawn-frame? spawn-frame? 1)
;(define make-splitter-frame
;  (lambda (behavior)
;    (vector 'splitter-frame behavior) ) )
;(defprimitive make-splitter-frame make-splitter-frame 1)
;(define splitter-frame?
;  (lambda (value)
;    (and (vector? value)
;         (eq? (vector-ref value 0) 'splitter-frame) ) ) )
;(defpredicate splitter-frame? splitter-frame? 1)
;(define make-wind-frame
;  (lambda (behavior prelude postlude)
;    (vector 'wind-frame behavior prelude postlude) ) )
;(defprimitive make-wind-frame make-wind-frame 3)
;(define wind-frame?
;  (lambda (value)
;    (and (vector? value)
;         (eq? (vector-ref value 0) 'wind-frame) ) ) )
;(defpredicate wind-frame? wind-frame? 1)
;(define wind-frame-prelude
;  (lambda (value)
;    (vector-ref value 2) ) )
;(defprimitive wind-frame-prelude wind-frame-prelude 1)
;(define wind-frame-postlude
;  (lambda (value)
;    (vector-ref value 3) ) )
;(defprimitive wind-frame-postlude wind-frame-postlude 1)
;(define make-mark
;  (lambda (q)
;    (vector 'mark q) ) )
;(defprimitive make-mark make-mark 1)
;(define mark?
;  (lambda (value)
;    (and (vector? value)
;         (eq? (vector-ref value 0) 'mark) ) ) )
;(defpredicate mark? mark? 1)
;(define mark-q 
;  (lambda (value)
;    (vector-ref value 1) ) )
;(defprimitive mark-q mark-q 1)

;;;                  [[ What if you do not have Meroon ]]
;;; Comment the following bunch of definitions.

(define-class frame Object (behavior))
(defprimitive frame-behavior frame-behavior 1)
(defprimitive make-frame make-frame 1)
(define-class prompt-frame frame ())
(defprimitive make-prompt-frame make-prompt-frame 1)
(defprimitive prompt-frame? prompt-frame? 1)
(define-class reset-frame frame ())
(defprimitive make-reset-frame make-reset-frame 1)
(defprimitive reset-frame? reset-frame? 1)
(define-class spawn-frame frame ())
(defprimitive make-spawn-frame make-spawn-frame 1)
(defprimitive spawn-frame? spawn-frame? 1)
(define-class splitter-frame frame ())
(defprimitive make-splitter-frame make-splitter-frame 1)
(defprimitive splitter-frame? splitter-frame? 1)
(define-class wind-frame frame (prelude postlude))
(defprimitive make-wind-frame make-wind-frame 3)
(defprimitive wind-frame-prelude wind-frame-prelude 1)
(defprimitive wind-frame-postlude wind-frame-postlude 1)
(defprimitive wind-frame? wind-frame? 1)

(define-class mark Object (q))
(defprimitive make-mark make-mark 1)
(defprimitive mark-q mark-q 1)  

(define (resume q v)
  (invoke (frame-behavior (car q)) (list (cdr q) v)) )

(define (extend q frame)
  (cons frame q) )

;;;==========================================================================
;;;==========================================================================
;;; Library of control operators

;;; These expressions are already written in ACPS style and will be
;;; evaluated before all. They are gathered in a different file (to be
;;; easily extracted by LiSP2TeX). 

;;; What follows is intended for interpreters since it mixes expansion-time
;;; and evaluation-time. If you want to compile this file, load it first
;;; so it can expand itself before being compiled.

(define *boot* '())

;;; Read a whole file

(define (file->list file)
  (call-with-input-file file
    (lambda (in)
      (let loop ((e (read in)))
        (if (eof-object? e)
            '()
            (cons e (loop (read in))) ) ) ) ) )

(define-pervasive-macro (interp-define)
  (set! *boot* '())
  (let* ((defs (file->list "hlco.scm"))
         (defs (map %expand-syntax defs)) )
    `(begin (set! *boot* (append! *boot* ',defs))
            ,@(map (lambda (def)
                     (case (car def)
                       ((set!) `(definitial ,(cadr def) 'void))
                       (else '42) ) )
                   defs )
            '*boot* ) ) )

(interp-define)

;;;==========================================================================
;;;==========================================================================
;;; Tests. This requires tester.scm to be present (since it defines
;;; interpreter and suite-test procedures).

;;;                       [[ Use the underlying expansion ]]
;;; You may remove all the previous expansion mechanism by just
;;; setting the %expand-syntax procedure to the right expander.

(set! %expand-syntax %expand-syntax)

(define empty-continuation "N'importe quoi!")

;;; A top-level interpreter: All errors must call the `wrong' procedure
;;; to restart the interpreter or suite-test.  That's why the global
;;; variable wrong is assigned to the synthesized err function.

(define (interp)
  (call/cc 
   (lambda (exit)
     (interpreter 
      "??  "
      "== "
      #t
      (lambda (read print err)
        (set! wrong err)
        ;; Only once for its side effect, it installs all the control operators.
        (evaluate `(begin . ,*boot*) r.global)
        (lambda () 
          (let ((e (read)))
            (if (eof-object? e)
                (exit 'done)
                (let* ((e (acps (%expand-syntax e) '() 'q.init))
                       (ne `((lambda (q.init) ,e) 
                             (list (make-frame (lambda (q value) 
                                                 value ))) )) )
                  (print (evaluate ne r.global)) ) ) ) ) ) ) ) ) )

;;; testing suite files.

(define (test-on-file file)
  (suite-test
   file
   "?? " "== " #t
   (lambda (read check err)
     (set! wrong err)
     ;; Once since used for its side effect
     (evaluate `(begin . ,*boot*) r.global)
     (lambda () 
       (let* ((e (acps (%expand-syntax (read)) '() 'q.init))
              (ne `((lambda (q.init) ,e) 
                    (list (make-frame (lambda (q value) value))) )) )
         ;;(pp ne) ;; DEBUG
         (check (evaluate ne r.global)) ) ) )
   special-equal? ) )

(define (special-equal? expected obtained)
  (cond ((and (pair? expected) (eq? (car expected) '*or*))
         (if *left-to-right*
             (special-equal? (cadr expected) obtained)    ; left->right
             (special-equal? (caddr expected) obtained)   ; right->left
             ) )
        ((equal? expected '?-) #t)
        ((pair? expected) 
         (and (pair? obtained)
              (special-equal? (car expected) (car obtained))
              (special-equal? (cdr expected) (cdr obtained)) ) )
        (else (equal? obtained expected)) ) )

;;; Runs all the test suites.

(define (test)
  (for-each (lambda (order)
              (set! *left-to-right* order)
              (let loop ((test-files 
                          '(
                            ;;"../../DEA/scheme-tests.scm"
                            ;;"../../scm/DECC/splitter.tests"
                            "contlib.tst"
                            )))
                (if (pair? test-files)
                    (begin
                      (newline)(newline)(newline)
                      (format #t ";;;    Test on ~A (evaluation order: ~A)"
                              (car test-files)
                              (if *left-to-right*
                                  "left to right"
                                  "right to left" ) )
                      (newline)(newline)(newline)
                      (and (test-on-file (car test-files))
                           (loop (cdr test-files)) )) ) ) )
            (list #t #f) ) 
  'done )

;;; end of contlib.scm
